#!/usr/bin/env python

from setuptools import setup

__version__ = '0.0.4'

CLASSIFIERS = map(str.strip,
"""Environment :: Console
Environment :: X11 Applications :: GTK
License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)
Natural Language :: English
Operating System :: POSIX :: Linux
Programming Language :: Python
Programming Language :: Python :: 2.7
Topic :: Security
""".splitlines())

entry_points = {
    'console_scripts': [
        'dns-firewall = dns_firewall.main:main',
    ]
}

setup(
    name="dns-firewall",
    version=__version__,
    author="Federico Ceratto",
    author_email="federico.ceratto@gmail.com",
    description="Local DNS Firewall",
    license="AGPLv3+",
    url="https://github.com/FedericoCeratto/dns-firewall",
    long_description="",
    classifiers=CLASSIFIERS,
    keywords="desktop security",
    install_requires=[
        'setproctitle>=1.0.1',
    ],
    packages=['dns_firewall'],
    package_dir={'dns_firewall': 'dns_firewall'},
    platforms=['Linux'],
    zip_safe=False,
    entry_points=entry_points,
    # Used by setup.py bdist to include files in the binary package
    package_data={'dns_firewall': [
        'data/*.png'
    ]},
)




"""
.. module:: dns_firewall.ui
   :synopsis: Local DNS firewall Gtk UI

"""

# Released under AGPLv3+ license, see LICENSE

from datetime import datetime
from gi.repository import Gtk, Gdk, GObject
from pkg_resources import resource_filename
from logging import getLogger
import gevent
import cairo
import os
import sys

Gtk.Widget.__enter__ = lambda self, *a, **kw: self
Gtk.Widget.__exit__ = lambda *a, **kw: None

ACTION_COLOR_MAP = dict(accept='green', drop='red', nxdomain='orange',
                        default='orange')
NUM_LOG_ROWS = 100

log = getLogger()


def _pos(menu, icon):
    return (Gtk.StatusIcon.position_menu(menu, icon))


def get_resource(fn):
    if 'DEVELOPMENT' in os.environ:
        p = 'dns_firewall'
    else:
        p = resource_filename(__name__, '.')

    ap = os.path.abspath(p)
    return os.path.join(ap, 'data', fn)


class PopupMenu(object):

    def __init__(self, icon):
        self._status_icon = icon
        with Gtk.Menu() as menu:
            with Gtk.MenuItem('Exit') as item:
                menu.append(item)
                item.connect('activate', self._status_icon.exit)

            menu.show_all()
            self._menu = menu

    def popup(self, widget, button, time):
        self._menu.popup(None, None, _pos, self._status_icon._icon, button,
                         time)


class LogsWindow(Gtk.Window):

    def __init__(self, trayicon):
        Gtk.Window.__init__(self, title="DNS firewall logs")
        self.set_style()
        self._trayicon = trayicon
        self.set_border_width(10)

        with Gtk.Box(spacing=6) as hbox:
            self.add(hbox)

            with Gtk.VBox(spacing=6) as vbox:
                hbox.add(vbox)

                with Gtk.Button('Close') as closebtn:
                    vbox.pack_start(closebtn, True, True, 0)
                    closebtn.connect('clicked', lambda _: self.hide())

                with Gtk.Button('Exit') as closebtn:
                    vbox.pack_start(closebtn, True, True, 0)
                    closebtn.connect('clicked', self._trayicon.exit)

                with Gtk.ScrolledWindow() as scrollw:
                    hbox.pack_start(scrollw, True, True, 0)
                    scrollw.set_size_request(250, 150)
                    self._scrolled_window = scrollw

                    with Gtk.Grid() as grid:
                        scrollw.add(grid)
                        self._logs_grid = grid

                vbox.connect('size-allocate', self._adjust_scrolling_window)

    def _adjust_scrolling_window(self, widget, event, data=None):
        adj = self._scrolled_window.get_vadjustment()
        adj.set_value(adj.get_upper() - adj.get_page_size())

    def toggle_visibility(self, *a):
        if self.is_visible():
            self.hide()
        else:
            self.show_all()

    def add_log_line(self, client_program_name, q_domain, domain_filter,
                     action):
        """Add a log line to the displayed list"""
        color = ACTION_COLOR_MAP.get(action, ACTION_COLOR_MAP['default'])
        tstamp = datetime.now().strftime('%H:%M:%S')
        row_num = len(self._logs_grid) / 4

        items = (tstamp, client_program_name, q_domain, action)
        for col_num, i in enumerate(items):
            with Gtk.Label(" %s " % i, xalign=0) as label:
                label.set_name("%s-background" % color)
                self._logs_grid.attach(label, col_num, row_num, 1, 1)

        if row_num > NUM_LOG_ROWS:
            self._logs_grid.remove_row(0)

        self._logs_grid.show_all()
        self._logs_grid.queue_draw()

    def set_style(self):
        style_provider = Gtk.CssProvider()

        css = """
        #red-background {
            background-color: #ffeeee;
        }
        #orange-background {
            background-color: #fff3d7;
        }
        #green-background {
            background-color: #eeffee;
        }
        """

        style_provider.load_from_data(css)

        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION
        )

    def _on_close(self, *a):
        self._window.close()


class TrayIcon(object):
    def __init__(self):

        self._menu_popup = PopupMenu(self)
        self._logs_window = LogsWindow(self)

        self._icon_cache = {}
        self._blinker_flags = {}
        self._icon = Gtk.StatusIcon()
        self._icon.connect('activate', self._logs_window.toggle_visibility)
        self._icon.connect('popup-menu', self._menu_popup.popup)
        self._icon.set_visible(True)
        self.set_icon()
        # self._logs_window.toggle_visibility()  # development

    def set_tooltip_text(self, text):
        self._icon.set_tooltip_text(text)

    def pos(self, menu, icon):
        return (Gtk.StatusIcon.position_menu(menu, icon))

    def set_icon(self):
        self._icon.set_from_pixbuf(self.generate_icon(self._blinker_flags))

    def set_blinker(self, color):
        try:
            GObject.source_remove(self._blinker_flags[color])
            redraw_needed = False
        except KeyError:
            redraw_needed = True

        t_id = GObject.timeout_add(300, self._reset_blinker, color)
        self._blinker_flags[color] = t_id
        if redraw_needed:
            self.set_icon()

    def _reset_blinker(self, color):
        try:
            GObject.source_remove(self._blinker_flags[color])
            del(self._blinker_flags[color])
        except KeyError:
            pass

        self.set_icon()

    def generate_icon(self, flags):
        """Generate tray icon based on the current set of active flags
        """
        flag_names = ('red', 'orange', 'green')
        flags = tuple(f in flags for f in flag_names)
        try:
            return self._icon_cache[flags]
        except KeyError:
            pass

        size = 64
        surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, size, size)
        ctx = cairo.Context(surface)

        bkg = cairo.ImageSurface.create_from_png(get_resource('bkg.png'))
        ctx.set_source_surface(bkg, 0, 0)
        ctx.paint()

        for status, name in zip(flags, flag_names):
            if status:
                fn = get_resource("%s.png" % name)
                img = cairo.ImageSurface.create_from_png(fn)
                ctx.set_source_surface(img, 0, 0)
                ctx.paint()

        pb = Gdk.pixbuf_get_from_surface(
            ctx.get_target(),
            0,
            0,
            ctx.get_target().get_width(),
            ctx.get_target().get_height()
        )
        self._icon_cache[flags] = pb
        return pb

    def run(self):

        GObject.timeout_add(100, self._idle, priority=GObject.PRIORITY_HIGH)

        while True:
            Gtk.main_iteration()
            gevent.sleep(.001)

    def _idle(self):
        gevent.sleep()
        return True

    def start(self):
        gevent.Greenlet.spawn(self.run)

    def exit(self, *a):
        """Close the applet"""
        sys.exit()

    def add_log_message(self, client_program_name, q_domain, domain_filter,
                        action):
        """Add log message to the logs window"""
        self._logs_window.add_log_line(client_program_name, q_domain,
                                       domain_filter, action)
        color = ACTION_COLOR_MAP.get(action, ACTION_COLOR_MAP['default'])
        GObject.timeout_add(1000, self.set_blinker, color)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
.. module:: dns_firewall
   :synopsis: Local DNS firewall

"""

# Released under AGPLv3+ license, see LICENSE

from fnmatch import fnmatch
from gevent.server import DatagramServer
from psutil._compat import lru_cache
from setproctitle import setproctitle
import argparse
import dpkt.dns
import logging
import psutil
import socket
import struct
import yaml

cfg = {}

log = logging.getLogger(__name__)

color_map = dict(
    accept=32,  # green
    drop=31,  # red
    nxdomain=31,  # red
    unmanaged=30,
    default=33,  # yellow, used by "return"
)


def setup_logging():
    log.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s %(levelname)s %(message)s')
    ch.setFormatter(formatter)
    log.addHandler(ch)


def color(action):
    col = color_map.get(action, color_map['default'])
    return "\033[0;%dm%s\033[1;0m" % (col, action)


def query_resolver(server, port, querydata):
    """Send the query to an external resolver
    """

    if cfg['udp_mode']:
        sendbuf = querydata
    else:
        # length
        Buflen = struct.pack('!h', len(querydata))
        sendbuf = Buflen + querydata

    data = None
    try:
        if not cfg['udp_mode']:
            s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        else:
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        # set socket timeout
        s.settimeout(cfg['socket_timeout'])
        s.connect((server, int(port)))
        s.send(sendbuf)
        data = s.recv(2048)
    except Exception as e:
        log.info('[ERROR] query_resolver: %s' % e.message)
    finally:
        if s:
            s.close()
        return data


def extract_domain(s):
    """Extract domain
    """
    domain = ''
    i = 0
    length = struct.unpack('!B', s[0:1])[0]

    while length != 0:
        i += 1
        domain += s[i:i + length]
        i += length
        length = struct.unpack('!B', s[i:i + 1])[0]
        if length != 0:
            domain += '.'

    return domain


def check_dns_packet(data, q_type):

    if len(data) < 12:
        return False

    if cfg['udp_mode']:
        flags = data[2:4]
    else:
        flags = data[4:6]

    if q_type == 0x0001:

        ip_len = data[-6:-4]
        answer_class = data[-12:-10]
        answer_type = data[-14:-12]

        test = (ip_len == '\x00\x04' and answer_class == '\x00\x01' and
                answer_type == '\x00\x01')

        if not test:
            return False

    reply_code = struct.unpack('>h', flags)[0] & 0x000F
    return reply_code == 0


def fetch_pid(port):
    assert isinstance(port, int)
    for c in psutil.net_connections(kind='udp'):
        # if not c.raddr and c.laddr == ('0.0.0.0', port):
        if c.raddr in ((), ('127.0.0.1', 53)) \
            and c.laddr in (('0.0.0.0', port), ('127.0.0.1', port)):
            return c.pid

    log.error('No pid found %d', port)
    for c in psutil.net_connections(kind='udp'):
        if c.pid:
            name = fetch_process_name(c.pid)
            log.error('repr %r name %s', c, name)
        else:
            log.error('repr %r', c)


def fetch_process_name(pid):
    fname = "/proc/%d/cmdline" % pid
    with open(fname) as f:
        return f.read().split('\0', 1)[0]


@lru_cache(maxsize=10, typed=True)
def identify_caller(address):
    ip_address, port = address
    pid = fetch_pid(port)
    if pid is None:
        return

    return fetch_process_name(pid)


def resolve_over_tor(name):
    """Resolve using a local Tor instance"""
    # Send Tor SOCKS RESOLVE request.
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    socks_addr = (cfg['tor_socks_ipaddr'], cfg['tor_socks_port'])
    s.connect(socks_addr)
    buf = struct.pack('>BBHI', 4, 0xF0, 0, 1) + '\x00%s\x00' % name
    s.send(buf)
    v, stat, port, ipaddr = struct.unpack('>BBH4s', s.recv(1024))
    ipaddr = socket.inet_ntoa(ipaddr)

    if stat == 90:
        return ipaddr

    log.debug("Failed resolution over Tor")


class DNSServer(DatagramServer):

    def handle(self, querydata, address):
        client_program_name = identify_caller(address)
        self._handle_query(client_program_name, querydata, address)

    def _forward_query_to_resolver(self, querydata, src):
        """Forward DNS query to resolvers"""
        key = querydata[2:].encode('hex')
        for item in cfg['udp_dns_servers']:
            ipaddr, port = item.split(':')

            response = query_resolver(ipaddr, port, querydata)
            if response is None:
                # or not check_dns_packet(response, q_type):
                continue

            if self.lru_cache is not None:
                self.lru_cache[key] = response

            self.socket.sendto(response, src)
            return

    def _decide_action(self, client_program_name, query):
        """Decide action"""
        # FIXME: clean this code up
        if query.qr == dpkt.dns.DNS_Q and len(query.qd) == 1 \
            and len(query.an) == 0 and len(query.ns) == 0:
            q = query.qd[0]
        else:
            return 'transparent'

        q_domain = q.name
        q_type = q.type

        if q_type not in (dpkt.dns.DNS_A, dpkt.dns.DNS_AAAA):
            return 'transparent'

        if q.cls != dpkt.dns.DNS_IN:
            return 'transparent'

        rules = cfg['filtering']
        filtering_rules = rules.get(client_program_name, []) + rules['general']

        for domain_filter, action in filtering_rules:
            if not fnmatch(q_domain, domain_filter):
                continue

            if self.tray_icon:
                self.tray_icon.add_log_message(
                    client_program_name,
                    q_domain,
                    domain_filter,
                    action
                )
            log.debug("%r %r is matching %r -> %s" % (client_program_name,
                q_domain, domain_filter, color(action)))
            return action


    def _return_nxdomain(self, query, src):
        """Return NXDOMAIN"""
        ans = dpkt.dns.DNS()
        ans.op = dpkt.dns.DNS_RA
        ans.rcode = dpkt.dns.DNS_RCODE_NXDOMAIN
        ans.qr = dpkt.dns.DNS_R
        ans.id = query.id
        ans.qd.append(query.qd[0])
        self.socket.sendto(ans.pack(), src)

    def _return_ipaddr(self, query, ip_addr, src, ttl=600):
        """Return a given ipaddr"""
        ans = dpkt.dns.DNS()
        ans.op = dpkt.dns.DNS_RA
        ans.rcode = dpkt.dns.DNS_RCODE_NOERR
        ans.qr = dpkt.dns.DNS_R
        ans.id = query.id
        ans.qd.append(query.qd[0])

        arr = dpkt.dns.DNS.RR()
        arr.cls = dpkt.dns.DNS_IN
        arr.type = dpkt.dns.DNS_A
        arr.name = query.qd[0].name
        arr.ttl = ttl
        arr.ip = socket.inet_aton(ip_addr)
        ans.an.append(arr)

        self.socket.sendto(ans.pack(), src)

    def _handle_query(self, client_program_name, querydata, src):
        """Handle incoming query, return response
        """
        query = dpkt.dns.DNS(querydata)
        action = self._decide_action(client_program_name, query)

        if action == 'accept':
            # forward query, use caching
            if cfg['use_tor']:
                ip_addr = resolve_over_tor(query.qd[0].name)
                if ip_addr:
                    self._return_ipaddr(query, ip_addr, src)

            else:
                self._forward_query_to_resolver(querydata, src)

        elif action == 'nxdomain':
            self._return_nxdomain(query, src)

        elif action.startswith('return '):
            # return a fixed IP address
            ip_addr = action[7:].strip()
            self._return_ipaddr(query, ip_addr, src)

        elif action == 'drop':
            # simply drop the query
            return





def start_dns_server(cfg, tray_icon):
    server = DNSServer('%s:%s' % (cfg["host"], cfg["port"]))
    if cfg['enable_lru_cache']:
        # FIXME: implement caching
        server.lru_cache = {}

    server.tray_icon = tray_icon
    server.serve_forever()


def start_tray_icon():
    from ui import TrayIcon
    tray_icon = TrayIcon()
    tray_icon.start()
    return tray_icon


def main():
    global cfg

    setproctitle('dnsfirewall')

    setup_logging()

    parser = argparse.ArgumentParser(description='TCP DNS Proxy')
    parser.add_argument(dest='config_fname', type=argparse.FileType('r'),
                        help='config file')
    parser.add_argument('--tray', help='Enable tray icon', action='store_true',
                        default=False)
    args = parser.parse_args()

    cfg = yaml.load(args.config_fname)

    if args.tray:
        tray_icon = start_tray_icon()
    else:
        tray_icon = None

    log.info("DNS servers: %s", ' '.join(cfg['udp_dns_servers']))
    log.info("Query timeout: %f", cfg['socket_timeout'])
    log.info("Enable cache: %r", cfg['enable_lru_cache'])

    start_dns_server(cfg, tray_icon)


if __name__ == "__main__":
    main()
